/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>

#define OWN
#include "dynamic.h"

#ifdef __HUGE__
void check(void) {
  int r=farheapcheck();
  if(r!=_HEAPOK) {
    fprintf(stderr,"\n heap problem\n");
    exit(1);
  }
}

void FAR * Calloc(size_t nitems,size_t size) {
  void *p;
  p = farcalloc(nitems, size);
  check();
  HeapFree=HeapMax=farcoreleft();
  return p;
}

void FAR * Malloc(size_t size) {
  void *p;
  p = farmalloc(size);
  check();
  HeapFree=HeapMax=farcoreleft();
  return p;
}

void FAR * Realloc(void *block, size_t size) {
  void *p;
  p = farrealloc(block,size);
  /* if (p) free(block); else p=block; */
  check();
  HeapFree=HeapMax=farcoreleft();
  return p;
}

#elif defined(IBMPC)

void check(void) {
  if(heapcheck()!=_HEAPOK) {
    fprintf(stderr,"\n heap problem\n");
    exit(1);
  }
}

void FAR * Calloc(size_t nitems,size_t size) {
  void *p;
  p = calloc(nitems, size);
  check();
  HeapFree=HeapMax=coreleft();
  return p;
}

void FAR * Malloc(size_t size) {
  void *p;
  p = malloc(size);
  check();
  HeapFree=HeapMax=coreleft();
  return p;
}

void FAR * Realloc(void *block, size_t size) {
  void *p;
  p = realloc(block,size);
  if (p) free(block);
  else p=block;
  check();
  HeapFree=HeapMax=coreleft();
  return p;
}

#else

#if defined(__APPLE__)
#include <malloc/malloc.h>
#else
#include <malloc.h>
#endif

void check(void) {;}

/*#define RETURN(a) void *rc=(a); if (!rc) ERROR0("not enough memory") else return rc*/
#define RETURN return
void FAR * Calloc(size_t nitems,size_t size) { RETURN(calloc(nitems, size)); }
void FAR * Malloc(size_t size) { RETURN(malloc(size)); }
void FAR * Realloc(void *block, size_t size) {RETURN(realloc(block,size));}

long int coreleft(void) { 
  /*struct mallinfo m=mallinfo();
  return( (long int)m.fsmblks + (long int)m.fordblks );*/
  return 1024*1024;
}

#endif
