/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* temporal mode of the response functions for the Cubic FitzHugh-Nagumo 
 * model in Winfree's notation [Chaos1(3):305,1991],
 * with some more parameters added
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "system.h"
#include "beatbox.h"

#include "device.h"
#include "state.h"
#include "bikt.h"
#include "rhs.h"
#include "qpp.h"

typedef struct {
  real
    a, b, w,  
    bet, gam,			/* Winfree's Kinetics A parameters */
    epsu, epsv;			/* may be required different */
  real 
    Iu,  Iv, Il, UU; 			/* External influence */
} STR;

RHS_HEAD(fhnresft,3) {
  DEVICE_CONST(real,a)
  DEVICE_CONST(real,b)
  DEVICE_CONST(real,bet)
  DEVICE_CONST(real,gam)
  DEVICE_CONST(real,epsu)
  DEVICE_CONST(real,epsv)
  DEVICE_CONST(real,Iu)
  DEVICE_CONST(real,Iv)
  DEVICE_CONST(real,Il)
  DEVICE_CONST(real,w)
  DEVICE_CONST(real,UU)
  real U, U2, V, l, dUdt;
  U = u[0];
  V = u[1];
  l = u[2];
  U2 = UU*UU;
  du[0] = U*(1-U2)/epsu + epsv*V + Iu - U*l;
  du[1] = -U/epsu - (epsv*gam + l)*V + Iv;
  du[2] = Il;  
} RHS_TAIL(fhnresft)

/*#include argsused.h*/
RHS_CREATE_HEAD(fhnresft)
  ACCEPTP(a,RNONE,RNONE,RNONE);
  ACCEPTP(b,RNONE,RNONE,RNONE);
  ACCEPTP(epsu,RNONE,RSUCC(0.),RNONE);
  ACCEPTP(epsv,RNONE,RSUCC(0.),RNONE);
  ACCEPTP(bet,RNONE,RSUCC(0.),RNONE);
  ACCEPTP(gam,RNONE,RSUCC(0.),RNONE);
  ACCEPTP(Iu,0,RNONE,RNONE);
  ACCEPTP(Iv,0,RNONE,RNONE);
  ACCEPTP(Il,0,RNONE,RNONE);
  ACCEPTP(w,0,RNONE,RNONE);
  ACCEPTP(UU,0,RNONE,RNONE);
RHS_CREATE_TAIL(fhnresft,3)
