/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* Hodkin & Huxley model (J Physiol 117:500-544,1952), 
 * m adiabatically excluded
 */
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "system.h"
#include "beatbox.h"
#include "HH.on"

#if HH

/* Temporarily: to mark undefined parameters */
#define UNKNOWN 0

#include "device.h"
#include "state.h"
#include "bikt.h"
#include "rhs.h"
#include "qpp.h"

static double cub(double x) {return x*x*x;}
static double qrt(double x) {double q=x*x; return q*q;}
/* Goldman(?) - H&H exponent */
static double hhx(double x){ if(-1.e-4<x && x<1.e-4) return 1+x/2.+x*x/12.; else return x/(1-exp(-x));}

typedef struct {
  #define _(n,v) real n;
  #include "hh.h"
  #undef _
  real I;	/* current/area = mV/ms*capacitance  - trmbr current */
  real IV;	/* mV/ms - intercellular current */
} STR;

RHS_HEAD(hh,3) {
  #define _(n,v) DEVICE_CONST(real,n)
  #include "hh.h"
  #undef _
  DEVICE_CONST(real,I)
  DEVICE_CONST(real,IV)
  real V = u[0];
  real h = u[1];
  real n = u[2];
  real *dV = du+0;
  real *dh = du+1;
  real *dn = du+2;
  real m, INa, IK, Il;

  #define alpm (km*hhx(-(V+25.)/10.))
  #define betm (km*4.*exp(V/18.))
  #define alph (kh*0.07*exp(V/20.))
  #define beth (kh*1./(exp((V+30)/10)+1))
  #define alpn (kn*0.1*hhx(-(V+10.)/10.))
  #define betn (kn*0.125*exp(V/80.))

  m=alpm/(alpm+betm);
  IK = gK*qrt(n)*(V-VK);
  INa = gNa*cub(m)*h*(V-VNa);
  Il = gl*(V-Vl);

  *dV = -1./C*( IK + INa + Il + I ) + IV;
  *dh = alph*(1-h)-beth*h;
  *dn = alpn*(1-n)-betn*n;

} RHS_TAIL(hh)

/*#include argsused.h*/
RHS_CREATE_HEAD(hh) {
  #define _(n,v) ACCEPTP(n,v,RNONE,RNONE);
  #include "hh.h"
  #undef _
  ACCEPTP(I,0,RNONE,RNONE);
  ACCEPTP(IV,0,RNONE,RNONE);
  MALLOC(*u,3L*sizeof(real));
  (*u)[0]=0;
  (*u)[0]=1;
  (*u)[0]=0;
} RHS_CREATE_TAIL(hh,3)

#endif
