/**
 *  Copyright (C) (2010-2013) Vadim Biktashev, Alexander Karpov, 
 *  Irina Biktasheva, Ross McFarlane, Sanjay Kharche,
 *  Mario Antonioletti, Adrian Jackson.
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* 
 * Execute a command in a subshell
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "system.h"
#include "beatbox.h"
#include "device.h"
#include "state.h"
#include "qpp.h"
#include "error.h"
/* #include "bikt.h" */
/* #include "windraw.h" */
/* #include "k_.h" */
/* #include "mpi_io_choice.h" */

typedef struct {
  int advance;
  char *echo;
  char *cmd;
  char code[1024];
  pp_fn compiled;
} STR;

void doit(char *cmd, pp_fn code, char *echo) {
  char localbuf[2048];
  int rc;
  if (cmd==NULL) return;
  if (*cmd=='\0') return;
  k_on();
  sprintf(localbuf,cmd,*(real *)execute(code));
  k_off();
  if (echo) {if (*echo) printf(echo,localbuf);}
  if (0!=(rc=system(localbuf))) MESSAGE("t=%d shell returned code %d\n",t,rc);
}

// /*#include argsused.h*/
RUN_HEAD(shell)
  DEVICE_ARRAY(char, cmd)
  DEVICE_CONST(pp_fn, compiled)
  DEVICE_ARRAY(char, echo)
  doit(cmd,compiled,echo);
RUN_TAIL(shell)

// /*#include argsused.h*/
// /*#include assused.h*/
DESTROY_HEAD(shell)
  FREE(S->cmd);
  FREE(S->echo);
DESTROY_TAIL(shell)

CREATE_HEAD(shell) {
  DEVICE_IS_SPACELESS
  DEVICE_MUST_BE_NOWHERE
  char l[2048];
  CALLOC(S->cmd,MAXSTRLEN,sizeof(char));
  CALLOC(S->echo,MAXSTRLEN,sizeof(char));
  ACCEPTI(advance,0,0,1);
  ACCEPTS(echo,"beatbox> %s\n");
  if (strlen(S->echo)) REALLOC(S->echo,strlen(S->echo)+1);
  ACCEPTS(cmd,"");
  ACCEPTS(code,"t");
  k_on();
  S->compiled=compile(S->code,deftb,t_real); CHK(S->code);
  k_off();
  if (strlen(S->cmd)) {
    REALLOC(S->cmd,strlen(S->cmd)+1);
  }
#if MPI
  /* To be on the safe side, we only run this device on one process. */
  if (mpi_rank == 0) {
    dev->s.runHere = 1;
    if (S->advance) doit(S->cmd,S->compiled,S->echo);
  } else {
    dev->s.runHere = 0;
  }
#else
  if (S->advance) doit(S->cmd,S->compiled,S->echo);
#endif
} CREATE_TAIL(shell,0)
